
#' Class for \code{Vario_Pi_search}. 
#' @details Records results for one subset \code{\{X1,...Xm\}} and one functional form \code{Y=f(X1,...Xm; varphi_1,...varphi_m)}. 
#' Search goes over all varphi_i to find them a partition Pi. Result is a Ranking of candidates Pi.
#' @field Data_E1_En,Dstd_E1_En Data,Data standardized; list: \code{E -> (X1,..Xm, Y)}
#' @field Cf_X_E1_En regression coefficients; list: \code{X_i ->} mat: \code{E_j, varphi_ik} at ik
#' @field lconf,rconf,pval as Cf_X_E1_En with resp. conf.interval,pval.
#' @field Pi_histories,C_histories subclasses, information about a single partition/coefficient
#' @field Pi_ranking main result
#' @field Pi_ranking_raw,Pi_ranking_nogroup,Pi_ranking_nscaled,Pi_ranking_correctone test versions
#' @field htest_invariant different hypothesis tests on whether varphi are invariant
#' @examples Classes overview:
#' cls_subset_search (one target Y, all subsets X)
#' > cls_pi_search (one subset, all partitions) 
#' > cls_pi (one partition, all coefficients) 
#' > cls_varphi (one coefficient)  
#' @export
cls_pi_search = setRefClass(
  "cls_pi_search",
  fields = list(
    Data_E1_En = "list", 
    Dstd_E1_En = "list", 
    Cf_X_E1_En = "list", 
    lconf = "list", 
    rconf = "list",
    pval = "list", 
    Rejects_X_E1_En = "list", 
    Fun_id = "character", 
    X_n = "numeric",
    X_C_n = "numeric",
    E_n = "numeric", 
    Pi_histories = "environment", 
    C_histories = "environment", 
    #
    # Pi search result
    Pi_ranking = "list",
    #including conf based on confint sampling frpm pi one ->inconf cases set to 1 (or better singleton groups?)
    # Versions
    Pi_ranking_raw = "list",
    #not including any correction for one group or confidence (kmax accurate, k=1 never guessed)
    Pi_ranking_nogroup = "list",
    #confidence based on confint f pi kmax. problem: should rather generate samples randomly (as we do for k=1) than equidist (as for kmax)
    Pi_ranking_nscaled = "list",
    Pi_ranking_correctone = "list",
    #hypoth testing for inv. case.
    #
    #Pi_ranking_lconf="list",
    Pi_kbest = "list",
    scores_kbest = "numeric", 
    # special case one group
    Pi_one = "list", 
    # score confidence
    confints = "list", 
    C_conf = "list", 
    # invariance
    htest_invariant = "list",
    invariant_conservative = "logical",
    invariant_locally="logical",
    #Regression
    regress_E1_En = "list",
    greedy_k = "logical",
    meta_computing = "logical",
    done = "logical",
    done_kbest = "logical",
    speedy = "logical",
    verbose = "logical",
    #residuals
    Pi_resids ="list",
    version_mdl = "logical",
    version_simp = "logical"
  )
)

# Entries -------
cls_pi_search$methods(
  init = function(Data_E1_En,
                  regr_result,
                  config) { 
    #Regression
    regress_E1_En <<-
      regr_result$regress_E1_En 
    htest_invariant <<- regr_result$htest_invariant
    invariant_conservative <<-
      !(FALSE %in% htest_invariant$chow_pairs)
    invariant_locally <<-
      (TRUE %in% htest_invariant$htest_pairs |TRUE %in% htest_invariant$chow_pairs | TRUE %in% htest_invariant$htest_coefs)
    #invariant_conservative <<- !(FALSE %in% htest_invariant$chow_pairs)
    
    Cf_X_E1_En <<- regr_result$cf_mats$cfs
    lconf <<- regr_result$cf_mats$lconf
    rconf <<- regr_result$cf_mats$rconf
    pval <<- regr_result$cf_mats$pval
    
    # Data Collection
    Data_E1_En <<- Data_E1_En
    Dstd_E1_En <<- standardize(Data_E1_En)
    
    #Inits
    X_n <<- length(Cf_X_E1_En)
    E_n <<- length(Data_E1_En)
    id = ""
    degs = c()
    for (mat in Cf_X_E1_En) {
      id = paste0(id, nrow(mat)) #col up to x^j -> id j
      degs = c(degs, nrow(mat))
    }
    Fun_id <<- id
    X_C_n <<- degs
    Pi_histories <<- create_h()
    C_histories <<- create_h()
    C_conf <<- list()
    greedy_k <<- config$greedy_k
    done <<- FALSE
    done_kbest <<- FALSE
    meta_computing <<- config$meta_computing
    verbose <<- config$verbose & !meta_computing
    version_mdl <<- config$version_mdl 
    version_simp <<- config$version_simp
    
    Pi_one_X_C = list()
    for(xi in 1:X_n)
      Pi_one_X_C[[xi]] = rep(0, X_C_n[xi]) #matrix(ncol = E_n, 
                         #    nrow = X_C_n[xi])
    Pi_one <<- list(E_part = rep(1, E_n), score_sum = 0, scores_X_C = Pi_one_X_C)
    
    # Precompute for one-group partition here
    # Unless this is alr stored in config, e.g. for mult runs for synth data
    if (!meta_computing) {
      if (!length(config$conf_precomputed))
        confints  <<-
          list(
            conf_pione = sample_conf_pi_one(E_n, it_n = 50, config$verbose, config$version_mdl),
            conf_nogroups = sample_conf_nogroups(E_n, it_n = 50, config$verbose, config$version_mdl)
          )
      else
        confints  <<- config$conf_precomputed
    }
  },
  
  
  
  # Called for one (Xi-Cj) pair (= alpha_i_j in E1 to En) when pi is seen in search 
  # seen before: add entry in existing Pi_history, ow: new Pi_history
  edit = function(E_part, 
                  C_part,
                  C_history){
    
    E_partnm = paste(E_part, collapse="")
    seen = (E_partnm %in% ls(Pi_histories))
    if(!seen){
      Pi_history = new("cls_Pi")
      Pi_history$init(E_part, X_n, X_C_n, version_mdl=version_mdl,version_simp=version_simp)
    } 
    else 
      Pi_history = Pi_histories[[E_partnm]]
    
    Pi_history$edit(C_history, C_part)
    Pi_histories[[E_partnm]] <<- Pi_history
    
    X_C_nm = paste0(C_history$xi, paste0("-", C_history$ci))
    if(!(X_C_nm %in% ls(C_histories)))
      C_histories[[X_C_nm]] <<- C_history
    
    # Pi_one
    Pi_one$scores_X_C[[C_history$xi]][C_history$ci] <<- C_history$mdl_one
  # Pi_one$score_sum <<- Pi_one$score_sum + C_history$mdl_one
    
    # If Partition Pi seen for all Xi, Cj
    if(Pi_history$done){
      Pi = list(E_part=E_part,
                Pi_history=Pi_history)
      Pi$score_sum=Pi_history$score_sum #needed?
      Pi$scoreE_sum=Pi_history$scoreE_sum
     
      Pi_ranking[[E_partnm]] <<- Pi
      Pi_ranking_nscaled[[E_partnm]] <<- Pi
      
     Pi_one$score_sum <<- sum(sapply(Pi_one$scores_X_C, function(entr) return(sum(entr))))
      
    }
    return(Pi_history) 
  }
)
    


cls_pi_search$methods(
  
  check_C = function(C_history){
    ci = C_history$ci
    xi = C_history$xi
    pis = ls(C_history$K_histories)
    for (P in pis)
    assert (P %in% ls(Pi_histories)
            & length(Pi_histories[[P]]$scores_X_C[[xi]][ci]) 
            & (Pi_histories[[P]]$scores_X_C[[xi]][ci]<1000) #todo Inf.
            & (Pi_histories[[P]]$scoresE_X_C[[xi]][ci]<1000), 
            paste(xi, paste(ci, paste(P))))
  },
  #check all ci,xi present 
  # todo omit f time
  check_pis = function () {

    for(P in ls(Pi_histories))
      sapply(1:X_n, function(xi)
        sapply(1:X_C_n[xi], function(ci){
          assert(length(Pi_histories[[P]]$scores_X_C[[xi]][[ci]])
                 & Pi_histories[[P]]$done
                 & Pi_histories[[P]]$scores_X_C[[xi]][[ci]] < 1000
                 & Pi_histories[[P]]$score_sum <Inf
                 & Pi_histories[[P]]$scoresE_X_C[[xi]][[ci]] < 1000
                 & Pi_histories[[P]]$scoreE_sum <Inf) 
          assert(P %in% ls(Pi_ranking))}))
  },
  
  finish_pis = function(){
    if(verbose) tic("(vario)       Postprocess")
    if(verbose) tic("(vario)       -Transfer")
    transfer_pis()
    if(verbose) toc()
    #if(!speedy){
    #  if(verbose) tic("(vario)       -Check")
    #  check_pis()
    #  if(verbose) toc()
    #}
    if(verbose)  tic("(vario)       -Sort")
    Pi_ranking <<- list.sort(as.list(Pi_ranking), score_sum) 
    Pi_ranking_nscaled <<- list.sort(as.list(Pi_ranking_nscaled), scoreE_sum)
    if(verbose) toc()
    if(verbose) tic("(vario)       -kbest")
    done <<- TRUE
    sel_kbest()
    if(verbose) toc()
    
    #Skip residuals
    cmp_resid()
    
    if(!meta_computing) {
      if(verbose) tic("(vario)       -conf & sort")
      add_confidence()
      if(verbose) toc()
    }
    
    if(verbose) toc()
  }
)


# Pi Ranking ---------
cls_pi_search$methods(
  
  # After search for each (Xi,cj), 
  #  transfer all seen partitions at one (X1, cf1) to other pairs (Xi, cj) 
  transfer_pis = function(){
    
    for(P in ls(Pi_histories)){
      Pi_history = Pi_histories[[P]]
      E_part = Pi_history$E_part
      E_partnm = paste(E_part, collapse = "")
      scores_X_C = Pi_history$scores_X_C
      
      for(xi in 1:length(scores_X_C)){
        scores_X = scores_X_C[[xi]]
        
        for(ci in 1:length(scores_X)){
          if(scores_X[ci]==1000){ #or Inf
            X_C_nm = paste0(xi, paste0("-", ci))
            assert(X_C_nm %in% ls(C_histories))
            C_history = C_histories[[X_C_nm]]
            C_part = reord_E_C_part(E_part, C_history)
            Pi_history$edit(C_history, C_part)
            Pi_histories[[E_partnm]] <<- Pi_history
          }
            if(Pi_history$done){
              Pi = list(E_part=E_part, 
                        Pi_history=Pi_history)  
              
              Pi$score_sum = Pi_history$score_sum
              Pi$scoreE_sum = Pi_history$scoreE_sum
                
              Pi_ranking[[E_partnm]] <<- Pi
              Pi_ranking_nscaled[[E_partnm]] <<- Pi
            }
                 
        }
      }
    }
  },
  
  sel_kbest = function(){
    assert(done)
    assert(length(ls(Pi_ranking)) == length(ls(Pi_histories)))
    k_seen = c()
    kbest = list()
    scores = c()
    
    for(pi in seq_along(Pi_ranking)){
      
      P = Pi_ranking[[pi]]$Pi_history
      k = length(unique(P$E_part))
      
      if(!(k %in% k_seen)) {
        kbest[[paste(k)]] = P
        k_seen = c(k_seen, k)
        scores= c(scores, P$score_sum)
      }
      
      if(!(FALSE %in% sapply(1:E_n, function(k) 
        return(k %in% k_seen)))) next
    }
    Pi_kbest <<- kbest
    scores_kbest <<- scores
    done_kbest <<- TRUE
  } #   TODO update kbest on the fly. omit if !precomputing
  
)



cls_pi_search$methods(
  
  # Score for Pi one 
  # based on confint over many runs
  add_confidence = function(){
    assert(done_kbest)
    best_k = as.numeric(names(Pi_kbest)[1])
    
    #Scores for each (Xi, Cj)
    Pio_X_C = list() #one 
    Pino_X_C = list() #none 
    
    score_sum = 0
    scoreE_sum = 0
    
    for(xi in 1:X_n){
      init =  matrix(ncol = E_n, 
                     nrow = X_C_n[xi])
      Pio_X_C[[xi]] = init
      Pino_X_C[[xi]] = init
      Rejects_X_E1_En[[xi]] <<- init 
      
      for(ci in 1:X_C_n[xi]){
        
        # only for display 
        cfs = Cf_X_E1_En[[xi]][ci,]
        ### print(paste("Cfs:", paste(out_br(clld(cfs, d=1)))))
        
        
        rank = Pi_ranking[[1]]$Pi_history
        best = rank$scores_X_C[[xi]][ci]
        
        best_n = Pi_ranking_nscaled[[1]]$Pi_history$scoresE_X_C[[xi]][ci]
        best_part =  Pi_ranking_nscaled[[1]]$E_part
        
        # Precomputed before: conf ints one-group pi
        #conf_pione = sample_conf_pi_one(E_n, cfs, numit = 50)
        
        #reject k>1 because score for k in confint for samples w k'=1
        conf_pione = confints$conf_pione
        conf_nogroups = confints$conf_nogroups
        
        C_conf[[ci]] <<- c(conf_pione=(conf_pione[1,best_k-1]<best), 
                           conf_nogroups=(conf_nogroups[1,best_k-1]<best))
        
        #conf = min(conf_nogroups[1,best_k-1], conf_pione[1,best_k-1]) #conf_pione  
        conf_one = conf_pione[1,best_k-1] # test  
        conf_none = conf_nogroups[1,best_k-1] # test
        
        
        if(best>conf_none) Rejects_X_E1_En[[xi]][ci,] <<- TRUE
        else Rejects_X_E1_En[[xi]][ci,] <<-  FALSE
        ##Vario_n_conf = conf_pione$Vario_n
        # add overall confidence:
        #...
        
        
        #print(paste0("     Cint k=",paste(best_k, paste(fmtd(conf_pione[1,best_k-1], d=2), 
        #                                               paste("<>", paste(fmtd(best,d=2), paste(", Conf nogroups", 
        #                                               paste( fmtd(conf_nogroups[1,best_k-1], d=2 ), paste("<>" ,fmtd( best ,d=2) ) ))    ))))))
        
        
        Pio_X_C[[xi]][ci,] = conf_one
        Pino_X_C[[xi]][ci,] = conf_none
        
        score_sum = score_sum + conf_one
        scoreE_sum = scoreE_sum + conf_none
      }
      Pio_score = score_sum
      Pino_score = scoreE_sum
      
      best = Pi_ranking[[1]]$score_sum
      #best_n = Pi_ranking_nscaled[[1]]$scoreE_sum
      
      # for completeness of entry
      Pio_X_C[[xi]] = sum( Pio_X_C[[xi]])
      Pino_X_C[[xi]] = sum( Pino_X_C[[xi]])
    }
    assert(length(unique(Pi_one$E_part))==1) #debug: pi one created by cmp resid
    
    # Entries 
    #Pi_one$scoreE_sum <<- scoreE_sum
    
  
 if (!version_mdl){ #ow handled during search
      Pi_one$score_sum <<- score_sum
      Pi_one$scores_X_C <<- Pio_X_C
 }
    E_partnm = paste(Pi_one$E_part, collapse="")
    #Pi_ranking_nscaled[[E_partnm]] <<- Pi_one
    
    # Pi_correctone not based on conf of score itself
    Pi_one_corrected = Pi_one
    if(TRUE %in% invariant_conservative){ #if invariance test succeds, force Pi_one to the top of ranking
      ##  print("Pione")
      ##  print(htest_invariant)
      Pi_one_corrected$score_sum  = Pi_ranking[[1]]$score_sum -1 #TODO hack? need a score here?
      
    } else 
      Pi_one_corrected$score_sum  = Pi_ranking[[length(Pi_ranking)]]$score_sum
    
    
    # Sort with score for Pi one included or not, and 2 score versions
    Pi_ranking_correctone <<- Pi_ranking
    Pi_ranking_correctone[[E_partnm]] <<- Pi_one_corrected
    Pi_ranking_correctone <<- list.sort(Pi_ranking_correctone, score_sum)
    
    Pi_ranking_nscaled <<- list.sort(as.list(Pi_ranking_nscaled), scoreE_sum)
    Pi_ranking <<- list.sort(as.list(Pi_ranking), score_sum)
    
    
    Pi_ranking_raw <<- Pi_ranking #not including pi one, original
    Pi_ranking_nogroup <<- Pi_ranking #not including pi one, overwrite
    
    Pi_ranking[[E_partnm]] <<- Pi_one
    Pi_ranking <<- list.sort(as.list(Pi_ranking), score_sum) 
    
    #Overwrite using confints of pi no groups
     if(!version_mdl){
       
     Pi_one$score_sum <<- scoreE_sum
      Pi_one$scores_X_C <<- Pino_X_C
      Pi_ranking_nogroup[[E_partnm]] <<- Pi_one
      Pi_ranking_nogroup <<- list.sort(as.list(Pi_ranking_nogroup), score_sum)  
     }
  }

)



# Not needed anymore: -------
# Regression Errors ---------
cls_pi_search$methods(
  
  # Compute residuals for each partition Pi
  # - errors made if mean coefficients in Pi's groups used in environments
  # - add one group Pi
  cmp_resid = function(){
    assert(done)
    
    #Start with one group
    pione = rep(1, E_n) 
    res_one = cmp_resid_P (pione) #resid_sum & E_resids (per environment)

    for(pi in seq_along(Pi_ranking)) 
      cmp_resid_P (Pi_ranking[[pi]]$Pi_history$E_part, 
                   res_one) #modifies Pi_resids, entering partitions in the same order as ranking
  },
  

  cmp_resid_P = function(E_part, 
                         res_one = list()){ #res_one not set if call for pione
    
    E_partnm = paste(E_part, collapse="")
    K = length(unique(E_part))
    
    # Residuals for Pi's groups
    # - each E: sum_X mean_ci * X^(|ci|), where mean_ci of E's group in Pi
    
    E_Yhat = lapply(1:E_n, function(i) return(0))
    E_Ystd = E_Yhat

    for(xi in 1:X_n)
      for(ci in 1:X_C_n[xi]){
        E_Yhat = cmp_resid_X_C(xi, ci, E_part, E_Yhat)
        E_Ystd = cmp_resid_X_C(xi, ci, E_part, E_Ystd, std=TRUE)
      }
      
    resid_sum = 0 
    restd_sum = 0
    E_resids = c()
    E_restds = c()
    E_scores = c() 
    E_scores_std = c() 
    
    for (ei in 1:E_n) {
      resid = sum((Data_E1_En[[ei]][["y"]]-E_Yhat[[ei]])^2)
      restd = sum((Dstd_E1_En[[ei]][["y"]]-E_Ystd[[ei]])^2)
  
      if(logg(resid)==0) stop()
      
     E_resids = c(E_resids, logg(resid))
      E_restds = c(E_restds, logg(restd))
   #  E_resids  = c(E_resids, resid)
  #   E_restds = c(E_restds, restd)
     
      if(K!=1) {
        E_scores = c(E_scores, (res_one$E_resids[ei]-resid)/(K-1))
        E_scores_std = c(E_scores_std, (res_one$E_restds[ei]-restd)/(K-1))
      }
      resid_sum = resid_sum + resid
      restd_sum = restd_sum + restd
    }
    resid_sum = logg(resid_sum)
    restd_sum = logg(restd_sum)
    #print(paste(ei, paste("-", resid_sum)))
    if(K!=1) score_sum =(res_one$resid_sum-resid_sum)/(K-1) #(resid_pione-resid_sum)/(K-1)
    if(K!=1) score_std =(res_one$restd_sum-restd_sum)/(K-1)  
   # if(K==1){ # do that in add_conf:
     # Pi_one = list(E_part=E_part,
      #                      E_resids=E_resids,
       #                     E_restds=E_restds,
        #                    resid_sum=resid_sum,
        #                    restd_sum=restd_sum,
         #                   rescore_sum=0,
          #                  rescore_std=0)
   # }
    if(K!=1)
      Pi_resids[[E_partnm]] <<- list(E_part=E_part,
                                     E_resids=E_resids, 
                                     E_restds=E_restds, 
                                     E_scores=E_scores, 
                                     E_scores_std=E_scores_std, 
                                     # needed f what?
                                     resid_one=res_one$resid_sum,
                                     restd_one=res_one$restd_sum,
                                     E_resids_one=res_one$E_resids, 
                                     #
                                     resid_sum=resid_sum,
                                     restd_sum=restd_sum,
                                     rescore_sum=score_sum,
                                     rescore_std=score_std)

    return(list(resid_sum=resid_sum,
                restd_sum=restd_sum,
                E_resids=E_resids,
                E_restds=E_restds)) #ret for pione
  },
  
  cmp_resid_X_C = function(xi, ci, E_part, E_Yhat, std=FALSE){

    X_C_nm = paste0(xi, paste0("-", ci))
    XI = paste0("x",xi)
    assert(X_C_nm %in% ls(C_histories))
    
    C_hst = C_histories[[X_C_nm]]
    E_cf = C_hst$E_cf
    #-> sorted by dataset order
    E_mns = infer_means(E_part, E_cf)
    
    for(ei in 1:length(Data_E1_En)) {
      #E_X = Data_E1_En[[ei]][[XI]]
      x_to_ci = paste0(XI, paste0("^", ci))
      if(ci==1) x_to_ci=paste(XI)
      #print(x_to_ci)
      
      if(std) {
        if(x_to_ci %in% ls(Dstd_E1_En[[ei]]))
          E_X_to_ci = Dstd_E1_En[[ei]][[x_to_ci]]
        else{
          E_X = Data_E1_En[[ei]][[XI]]
          E_X_to_ci = E_X^ci
      }}
       else{
         if(x_to_ci %in% ls(Data_E1_En[[ei]]))
           E_X_to_ci = Data_E1_En[[ei]][[x_to_ci]]
         else{
           E_X = Dstd_E1_En[[ei]][[XI]]
           E_X_to_ci = E_X^ci
         }
       }
      
      E_class = E_part[ei]
      E_mn = E_mns[E_class]
       
      Yhat_exc =  E_mn * E_X_to_ci
        # this is (E_X^ci) 
      #sum this over all X and C to describe Y in environment E 
      E_Yhat[[ei]] = E_Yhat[[ei]] + Yhat_exc 
    }  
    return(E_Yhat) 
  })
  
  