
#' Check whether Vario can find the partition Pi, case: parameters change across groups, \code{Y=f^{Pi}(X) + N = alpha^{Pi}X + N
#' @details soft intervention or mechanism change at Y in the groups of Pi, as opposed to noise interventions.
#' Generate a random DAG, select nodes Y, generate data for Y using pa_G(Y) and group weights in each context, generate remaining nodes and run Vario
#' @param E_n num contexts
#' @param D_n num datasets
#' @param subset_search T: pa_G(Y) is unknown, run Vario_subset_search on complete graph; F: run Vario_Pi_search with known parents to make sure Pi is found there
#' @param X_max_n in case subset_search=T, max. subset size in Varios subset search
#' @param intervene_pa_G interval: in how many contexts pa(Y) is intervened on
#' @export
synth_DAG_param = function(
  E_n=5,
  D_n=1000,
  subset_search=TRUE,
  interventional_vs_observational=TRUE, 
  one_intervention_per_context=TRUE,
  perfect_interventions=FALSE, # in the 2nd group, interventions set coefficients to zero. other groups generated as before.
  X_max_n=4, 
  intervene_pa_G=c(3,3),  
  node_n=10,  
  deg_p=0.3, 
  it_pre=50,  
  #For inter, we can vary the distances between groups to add uncertainty; if inter=c(1,1), groups are equidistant, outherwise can keep inter[1]=1 and choose arbitrary inter[2]
  inter=c(1,2), #inter=c(1,1) for perfect samples 
  noise = c(0,0.05), #or c(0,0) for perfect samples
  #it_shape=1, 
  confints_mdl=NULL,
  confints_basic=NULL,
  show_context_coefficients=FALSE,
  vb=TRUE,
  version_mdl = TRUE,
  version_simp = TRUE,
  greedy_k = FALSE, # pi search
  subdir_nm="", #if we want to specify further info abt a run
  seed=100,
  up_to_groupnumber_k=5
  ){
  it=0
  if(interventional_vs_observational) mx_n_groups=2
  else  mx_n_groups=E_n
  # File write
   if(subset_search) out_dir="~/Vario/out/check_causal/"
   else out_dir="~/Vario/out/check_pi/"
   
   out_createdir(out_dir)
   
   if(version_mdl) out_dir = paste0(out_dir, "mdl/")
   else  out_dir = paste0(out_dir, "basic/")   

   out_createdir(out_dir) 
   
   
   out_dir = paste0(out_dir, subdir_nm)
   out_createdir(out_dir)
   
   out_dir = paste0(out_dir, "D", E_n, "_N", noise[2], "_")
  
  # Init counters of accuracy on Pi and causal parents
  if(!subset_search) counter = c(0,0,0,0,0)  # pi T, pi F, pi F but pi*<pi(found less precise groups),pi<pi*(found more finegrained groups), pi F and inconsistent (wrongly assigned datasets) 
  else counter = c(0,0,0,0,0, 0,0,0) #the above + parent TP, parent FP, parent FN 
  records = list()
  
  # Pi for children of Y is usually chosen  at random
  shapes_ch = sample_shapes(E_n, 50, incl_one_group = FALSE, incl_no_group = TRUE)
  
  # Sample Pi (as a shape, permutations irrelevant)
  if(mx_n_groups<E_n)
    # All partitions w up to mx_n_groups many groups
    shapes = sample_shapes_K(E_n, 50, mx_n_groups)
    else shapes = shapes_ch 
  
  if(one_intervention_per_context)
    shapes= list(c(E_n-1,1)) #diagonal design->intervene in the last context.
  
  # Precompute 
  if(!length(confints_basic)) 
    confints_basic = list(conf_pione=sample_conf_pi_one(E_n, it_pre, vb=FALSE, version_mdl=FALSE),
                        conf_nogroups= sample_conf_nogroups(E_n, it_pre, vb=FALSE, version_mdl=FALSE),
                        conf_best=sample_conf_pi_best(E_n, it_pre, vb=FALSE, version_mdl=FALSE))
  if(!length(confints_mdl)) 
    confints_mdl = list(conf_pione=sample_conf_pi_one(E_n, it_pre, vb=FALSE, version_mdl=TRUE),
                   conf_nogroups= sample_conf_nogroups(E_n, it_pre, vb=FALSE, version_mdl=TRUE),
                   conf_best=sample_conf_pi_best(E_n, it_pre, vb=FALSE, version_mdl=TRUE))
  
  config_mdl = init_config_analysis(confints_mdl, greedy_k=greedy_k, version_mdl = TRUE, version_simp=version_simp) 
  config_basic = init_config_analysis(confints_basic, greedy_k=greedy_k, version_mdl = FALSE,version_simp=FALSE) 
  

  # Test each Pi
  for(shape in shapes){  
    if(length(shape)>up_to_groupnumber_k) next
    
    if(interventional_vs_observational)
      if((shape[1])==1) next #skip the case where theres just one observational context for Y at all- otherwise cannot run ICP to find invariance.
    records[[paste(shape,collapse="")]] = list() 
    
    # - random partition for children:
    shape_ch = shapes_ch[[sample.int(length(shapes_ch), 1)]]
    
    # - in case one_intervention_per_context, generate_DAG_datasets should use invariant coefficients in all but one context
    if(one_intervention_per_context) shape_ch = c(E_n)
    
    # - More finegrained pi for children than for parents: 
    # shape_ch = c()
    # ctx_i = 1
    #for (grp in shape){
    #  shape_ch = c(shape_ch, 1)
    # if(grp>1) shape_ch = c(shape_ch, grp-1)
    #ctx_i = ctx_i + grp 
    # }
      
    # Generate a graph G for Pi ----
    dag = generate_DAG_datasets(E_n, D_n, 
                              shape, shape_ch, 
                              interv_future = TRUE, # only relevant for domain adapt
                              interv_X_n = intervene_pa_G, 
                              inter=inter, 
                              noise=noise,
                              node_n=node_n, 
                              seed=seed,
                              deg_p=deg_p,
                              #DAG info:
                              vb=subset_search, #omit dag info for partition search
                              one_intervention_per_context=one_intervention_per_context,
                              perfect_interventions=perfect_interventions, 
                              subdir_nm = subdir_nm,
                              show_context_coefficients=show_context_coefficients)
    
    # If this seed generates a DAG where each node has empty Markov Blanket
    if(dag$empty_dag) break
    nodes = 1:length(dag$node_MB)
    
    for (mdl in c(version_mdl)){
      cat("************************\n")
      if(mdl) cat("Vario with MDL\n")
      else cat("Vario basic\n")
      cat("************************\n")
      
      
      if(mdl) config = config_mdl
      else config = config_basic
      if(mdl) confints = confints_mdl
      else confints = confints_basic
 
      
    if(subset_search & vb)  { 
      cat("------------------------\n")
      cat("Checking Pi and pa_G(Y) \n")
       cat("------------------------\n")
    }
    if(!subset_search & vb){
      cat("------------------------\n")
      cat("Checking Pi with pa_G(Y) known \n")
      cat("------------------------\n")}
    
    # Target all nodes -------
    for(node_i in nodes){
      #if we do not want variables without parents: 
      if(length(dag$node_MB[[node_i]]$pa_ind)<1) next # or <2
      
      EXAMPLE_DAG<<-dag
      
      #Special case no parents
      # - in Pi_search, let Vario predict (wo. subset search, solely based on score for one set) from children (to make it harder to predict no-group or insignificant)
      # - or, if no children, from all vars.
      # -> Goal: Vario shouldnt be confident.
      if(length(dag$node_MB[[node_i]]$ch_ind))
        subset_if_no_pa = dag$node_MB[[node_i]]$ch_ind
      else subset_if_no_pa = 1:5 
      
      #Find Pi + parents  ----
     if(subset_search) {  
       
        result = synth_check_subset_search(data=dag$node_data[[node_i]], 
                                           X_max_n=X_max_n,
                                           shape_to_part(shape), 
                                           markov_blanket=dag$node_MB[[node_i]],
                                           confints=confints, out_nm =out_dir,
                                           seed=seed, node_i=node_i, vb=vb, version_mdl=mdl)
     }
     # Find Pi | parents ----- 
      else {

        result = synth_check_pi(data=dag$node_data[[node_i]], shape, 
                                num_pa=length(dag$node_MB[[node_i]]$pa_ind), subset_if_no_pa, config,
                                out_nm=out_dir,seed)
          
      } 
    } 
    } 
  } 
} 

#' Check Pi* < Pi, each group in Pi* is a group in Pi
#' @param shape Pi* as seq. of group lengths
#' @param pi Pi as seq. of group IDs
#' @export
check_consistent_shape_pi=function(shape, pi){
 pi_consistent = TRUE
 id = 1 
 for(grp in shape){
 if(grp>1){
 grp_in_guess = pi[id:(id+grp-1)]
 pi_consistent = pi_consistent & (length(unique(grp_in_guess))==1)
     }
 id = id+grp
  }
 return(pi_consistent)
 }

# 
# for (seed in 101:150){
#   res = synth_DAG_param(confints=confints, E_n=5, D_n=1000,
#                         subset_search=FALSE,
#                         X_max_n=4,
#                         invariant_ch_G=FALSE,
#                         partition_ch_G=FALSE,
#                         intervene_ch_G=TRUE,
#                         intervene_pa_G=c(5,5),
#                         node_n=10, #15,
#                         deg_p=0.5,
#                         it_pre=50,
#                         it_shape=1,
#                         out_dir="~/Vario/out_checkpi_Vario/",
#                         seed=seed)}
# for(seed in 150:200) res = synth_DAG_param(confints=confints, E_n=5, D_n=1000,
#                                            subset_search=FALSE,
#                                            X_max_n=4,
#                                            invariant_ch_G=FALSE,
#                                            partition_ch_G=FALSE,
#                                            intervene_ch_G=TRUE,
#                                            intervene_pa_G=c(5,5),
#                                            node_n=10, #15,
#                                            deg_p=0.3,
#                                            it_pre=50,
#                                            it_shape=1,
#                                            out_dir="~/Vario/out_checkpi_Vario/",
#                                            seed=seed)
