
generate_DAG = function(node_n, deg_p){
dag =  pcalg::randomDAG(node_n, deg_p,  lB = 1, uB=1) #, lB = 0.1, uB = 1)
adj = as(dag, "matrix")
 return(adj)
}
 
#' Generates DAG and for each target variable Y in it: a dataset with samples in all contexts, where Y=f^{Pi}(X) and X are Gaussian noise variables 
#' @param E_n num contexts
#' @param D_n num samples/context
#' @param shape partition pi (no gaps) for f^{pi}: pa(Y)->Y. length = num groups, entry=size groups. 
#' @param interv_Y_ty intervention type in all but the first group (observational). 1 means parameter changes, 2 means shift intervention.
#' @param interv_X_n interval: how many interventions on X over contexts
#' @param node_n DAG num nodes
#' @param deg_p DAG out prb
#' @example Ex.: shape=c(3,2) and interv_ty=1 could lead to X1s coefficients c(0.25,0.27,0.3,0.8,0.9) over 5 contexts;
#' shape=c(3,2) and interv_ty=1 can lead to invariant coefficients for X1 but noise interventions in group 2.
#' @export
generate_DAG_datasets = function(E_n=5, D_n=1000,  
                                shape = c(4,1), #means one soft intervention on Y    
                                shape_ch = c(1,1,1,1,1),   
                                #shift interventions:
                                interv_X_n =c(3,5), #  3 to 5 noise interventions on variables X, in 0 to 3 random contexts 
                                perfect_interventions=TRUE, #in 2nd group, set coefficients to zero -- meant for experiments with n.groups=2
                                interv_future = FALSE, #the last context is intervened upon
                                inter=c(1,2), #vary inter-group distance iff imin is not imax 
                                noise=c(0,0.1),
                                node_n=10, 
                                deg_p=0.3,
                                seed=100,
                                subdir_nm="Example",
                                #out_dir="~/Vario/out/graphs_true/",
                                show_context_coefficients=FALSE,
                                one_intervention_per_context=TRUE, #to make it easier for JCI?
                                skip_no_parent_nodes=FALSE, #to save time -- for example for perfect interventions, we can expect that children are always found if there are no parents, so no use in trying. 
                                vb = TRUE,
                                write_data=FALSE
                                #write_graphs=FALSE
                                ){
 # These dirs will be used:
  if(write_data){
  out_createdir(paste0("~/Vario/data_synthetic/", subdir_nm))
  out_createdir(paste0("~/Vario/out/rdata_vario/", subdir_nm))
  out_createdir(paste0("~/Vario/out/check_causal/", subdir_nm))
  out_createdir(paste0("~/Vario/out/graphs_true/", subdir_nm))
  out_createdir(paste0("~/Vario/out_relatedwork/", subdir_nm))
  } 
  set.seed(seed)
  # DAG 
  adj = generate_DAG(node_n, deg_p)
  if(vb) cat("------------------------\n")
  if(vb) cat("Generating DAG,", node_n, "nodes", "\n")
 # if(vb) sink(stdout(),paste0("Generating DAG,", node_n, "nodes", "\n"))
  if(vb) cat("  -Seed:", seed, "\n")
  if(vb) cat("  -Targets:", 1:node_n, "\n")
  if(vb) cat("  -Pi(Y), group sizes:",paste(shape, collapse=","),"\n") 
  if(!one_intervention_per_context){
  if(vb) cat("  -Pi(children_G(Y)):",paste(shape_ch, collapse=","),"\n") 
    if(vb) cat("  -Perfect interventions Y: ", perfect_interventions, "(in the 2nd group)", "\n")
    if(vb) cat("  -Distribution shifts of pa_G(Y): noise shift in", interv_X_n[1], "to", interv_X_n[2], "contexts\n")
    if(vb) cat("  -Distrib.shift of pa_G(Y) in last (future) context: ", interv_future, "\n")
    
  }
  else
    if(vb)  cat("  -For each context, one intervention on a system variable,\n" ) 
   
  # Each target Y in DAG -----
  # system vars in E_n contexts
  node_data = list()
  # system + context vars, diagonal design
  node_data_jci = list() 
  node_MB = list()
  node_adj = list()
  node_i = 0
  # node_i may differ from y_i in case we skip nodes
  
  for (y_i in 1:node_n){
    node_i = node_i + 1 
    data_i = list()
    
    #Consider the Markov Blanket around Y
    MB = get_MB(adj, y_i)
    # Skip if the MB is empty; if this happens for all nodes, we skip the DAG later.
    if(length(MB$mb)==0) next
    # Skip also if nodes without parents should be ignored
    if(skip_no_parent_nodes & length(MB$pa_ind)==0) next
    
    # If needed, choose for each context one MB-variable that we intervene upon.
    # By convention, the last context theres an intervention on Y itself.
    non_interv_targets = MB$mb
    interv_targets = c()
    if(one_intervention_per_context){
      for(ci in 2:(E_n-1)) {
        if(length(non_interv_targets)==0) #already used up all contexts for too many system variables
          break
        chosen = non_interv_targets[sample.int(length(non_interv_targets),1)]
        non_interv_targets=non_interv_targets[-which(non_interv_targets==chosen)]
        interv_targets = c(interv_targets, chosen)
      }
      
      if(vb)  cat("  -Intervention targets:", paste(interv_targets, collapse = ","),  "\n" ) 
      
    }
    X_n = length(MB$mb) #Covariates 
    Xy_n = X_n + 1 #Target
    V_n = X_n + 1 + E_n - 1 # + ContextV. for all but one observational context
    X_nms = paste0("x", 1:X_n)
    
    #Init Data ---------
    data_jci_i = matrix(0, ncol=V_n, nrow=E_n* D_n)  
     #Init context data in vario format, for all MB-variables incl. Y
    for(context in 1:E_n) {
      data_i[[paste0("D",context)]] = data.frame(matrix(0, nrow=D_n, ncol=Xy_n))     
      
      #names for display
     if(length(X_n)) names(data_i[[paste0("D",context)]]) = c(X_nms, "y")
     else names(data_i[[paste0("D",context)]]) = c("y")
    }
    # init JCI data in diag.design
    for(context in 2:E_n) {
      context_rows = c(((context-1)*D_n+1):(context *D_n))
      data_jci_i[,X_n + 1 + context-1][context_rows]  = rep(1, D_n)
    }
    
    
    # 1. Non-children of Y ----
    # - Gaussian variables in each context  
    # - simple shift interventions in a number of noisy contexts, aswe need not care about the exact partition
    # - number of noisy contexts: sampled from interventions_X_number, contexts chosen at random
    

    if(MB$non_ch_n>0)  
      #traverse reordered indices 
      for(x_j in MB$non_ch_ind){ 
        x_nm = paste0("x", x_j)
        
        # Choose contexts where X has distribution shifts (similar to a partition)
        if(interv_X_n[1] == interv_X_n[2]) noisy_n = interv_X_n[1]
        else noisy_n = sample.int( (interv_X_n[2]-interv_X_n[1]), 1) + interv_X_n[1]
        assert(noisy_n >=0 & noisy_n <=E_n)
        noisy_contexts = c()
        if(noisy_n >0){
          if(noisy_n< E_n)
          noisy_contexts = combn(E_n, noisy_n)[,sample.int(choose(E_n, noisy_n),1)] 
          else noisy_contexts = 1:E_n
        }
        
        
        
        # Make sure that the last context is intervened upon, and 2nd to last is not intervened upon
        # case: want to predict labels Y in such context
        if (interv_future) {
          noisy_contexts = c(2,4,5,6,7)
          #if(!(1 %in% noisy_contexts)) 
          #  noisy_contexts = c(noisy_contexts, 1) 
          #if(2 %in% noisy_contexts)
          #  noisy_contexts = noisy_contexts[-which(noisy_contexts==2)]
        }
        # Exception: 
        if(one_intervention_per_context) noisy_contexts = c()

        
        mn = 0
        for(context in 1:E_n) {
          if(context %in% noisy_contexts) mn = 1
          if(context %in% noisy_contexts) scale = 2
          else scale =1  
          if(context %in% interv_targets) X_data= scale*rnorm(D_n, mean=1) 
            else X_data =scale* rnorm(D_n, mean=mn) 
          # Data as context list, only system variables
          data_i[[paste0("D",context)]][[x_nm]] =  X_data
          
          # Data in diagonal design, with context variables
          context_rows = c(((context-1)*D_n+1):(context *D_n))
          data_jci_i[,x_j+1][context_rows] =  scale* rnorm(D_n, mean=mn)
        } 
      }
    
    
    
    #TDOOO continue w interv_targets.
    
    # 2. Y generated according to Pi ------
    # -where pi indicates groups 
    # -no parents: Y is gaussian variable
    if(length(MB$pa_ind) == 0){
      if(show_context_coefficients) cat("\nLinear coefficients Y", node_i, ", " ,E_n, " contexts: zero parents. \n", sep="")
      
      for(context in 1:E_n) {
        data_i[[paste0("D",context)]][["y"]] =  rnorm(D_n)  
        
        # Data in diagonal design, with context variables
        context_rows = c(((context-1)*D_n+1):(context *D_n))
        data_jci_i[,1][context_rows] =   rnorm(D_n) 
      }
      if(one_intervention_per_context){
        context_rows = c(((context-1)*D_n+1):(context *D_n))
        data_i[[paste0("D",context)]][["y"]] =  rnorm(D_n, mean=1)   
        data_jci_i[,1][context_rows] =   rnorm(D_n, mean=1) 
      }
    }
    else {
      # Weight matrix 
      
      if(show_context_coefficients) cat("\nLinear coefficients Y", node_i, ", " ,E_n, " contexts, " , length(MB$pa_ind) , " parent(s), perf. interv. in 2nd group ", perfect_interventions,"\n", sep="")
      coef_X_C = gen_DAG_weights_Pi(length(MB$pa_ind), 
                                    shape, inter, noise, vb=show_context_coefficients,
                                    perfect_interventions=perfect_interventions)
      #cat("\t Coef", coef_X_C)
      
      # Convention: The first context group is observational.
      # The other contexts are generated as some modification.
      for(context in 1:E_n)  { 
        coef_X = coef_X_C[context,]
           X = data_i[[paste0("D",context)]][MB$pa_ind]
           
        assert(nrow(X)==D_n)
        assert(ncol(X)==length(MB$pa_ind))
        
        if(length(MB$pa_ind)>1){
          Y = as.matrix(X)%*%diag(coef_X)
          Y = apply(Y, 1, sum) + rnorm(D_n)
        }
        else{
          Y = as.matrix(X)%*%coef_X
          Y = Y[,1] + rnorm(D_n)
        }
        data_i[[paste0("D",context)]][["y"]] = Y
        
        #Diagonal data
        context_rows = c(((context-1)*D_n+1):(context *D_n))
        data_jci_i[,1][context_rows]=Y 
      }
    }
    
     
    #3. Children of Y -----
    # - as opposed to non-children, we generate from 1. linear eq. instead of noise variable, 2. with partition instead of some interventions, since we need to make sure Vario does not find the reverse equation.
    ch_ct = 0
    for (x_j in MB$ch_ind){
      
      x_orig = MB$mb[x_j]
      ch_ct = ch_ct +1
      # Parents 
      pa_orig = which(adj[,x_orig]==1)
      pa_j = which(MB$mb %in% pa_orig) # excludes Y and takes index order as in MB. 
      
      # Distribution shifts/non-iidness: for each group for Y, choose at least one context where X is noniid

      noisy_contexts = c()
      if (interv_future) {
        
        noisy_contexts = c(2,4,5,6,7)
        #if(!(1 %in% noisy_contexts)) 
        #  noisy_contexts = c(noisy_contexts, 1) 
        #if(2 %in% noisy_contexts)
        #  noisy_contexts = noisy_contexts[-which(noisy_contexts==2)]
      }
        coef_X_C = gen_DAG_weights_Pi(length(pa_j)+1, # Y itself is also a parent so +1.
                                      shape_ch, inter, noise,
                                      vb=FALSE, perfect_interventions=FALSE)
     
        for(context in 1:E_n)  { 
          mn = 0
          if(context %in% noisy_contexts) mn = 1
          if(context %in% noisy_contexts) scale = 2
          else scale = 1
          #else mn = 0
         
          coef_X = coef_X_C[context,] 
          if(one_intervention_per_context) {
            mn = 1
            coef_X = rep(0, length(coef_X))
          }
          
          if(length(pa_j) == 0)
            X = data_i[[paste0("D",context)]]["y"]

         else  X = cbind(data_i[[paste0("D",context)]][pa_j], 
                    data_i[[paste0("D",context)]]["y"])
          
          assert(nrow(X)==D_n)
          assert(ncol(X)==length(pa_j)+1)
          
          if(length(pa_j)>0){
            Y = as.matrix(X)%*%diag(coef_X)
            Y = apply(Y, 1, sum) + scale * rnorm(D_n, mean=mn)
          }
          else{
            Y = as.matrix(X)%*%coef_X
           Y = Y[,1] + scale * rnorm(D_n, mean=mn)
          }
          data_i[[paste0("D",context)]][[paste0("x",x_j)]] = Y
          context_rows = c(((context-1)*D_n+1):(context *D_n))
          data_jci_i[,x_j+1][context_rows] =  Y 
        } 
    }
    
    node_data[[paste0("y", node_i)]] = data_i
    node_data_jci[[paste0("y", node_i)]] = data_jci_i
    node_MB[[paste0("y", node_i)]] = MB
  }
  
  
  if(!length(node_data)) {#every MB was empty -- skip this seed.
    
    if(vb) cat("  -EMPTY DAG, skipping this seed.\n")
    stop()
    return(list(empty_dag = TRUE))
  }
  # create the augmented adjacency matrix with an intervention node for each context
  context_nodes = (node_n+1):(node_n+E_n-1) #first one is observational 
  aug_node_n = node_n + E_n-1
  aug_adj = matrix(0,aug_node_n,aug_node_n)
  for(row_i in 1:nrow(adj)){
    for(col_j in 1:ncol(adj)){
      aug_adj[row_i, col_j] = adj[row_i, col_j]
    }
  } # no adjacency to context vars for this graph yet, data and partitions chosen for each target later so consider by-target graphs
  
 #  if(!write_graphs & !write_data) 
  #   return(list(node_data=node_data,
   #              node_MB=node_MB, 
    #             adj=adj))
 
  #TODO: X has one obsv plus En-1 context varis write.csv(X,file=paste(out_nm,'data.csv',sep=''),row.names=FALSE)

    
  for(node_i in 1:length(node_data)){
    
    mb = node_MB[[paste0("y", node_i)]]$mb  
    
    
    node_n_i = length(mb) + 1
    aug_node_n_i = node_n_i + E_n-1
    context_nodes_i = (node_n_i+1:aug_node_n_i)
    
    aug_adj_i = matrix(0,aug_node_n_i,aug_node_n_i)

    #REwrite adjacency matrix, restricted to Markov Blanket around Y and with the new variable ordering
    # - MB leads to unique ordering(target Y at index 1, then first its parents, then children, then pa-of-ch, otherwise the original ordering )
    # - all edges not going to Y are irrelevant for us for now
    
    for(row_i in 1:node_n_i){
      for(col_j in 1:node_n_i){
        if(row_i==1) orig_pa=node_i
        else
          orig_pa = mb[row_i-1]
        if(col_j==1) orig_ch=node_i
        else
          orig_ch = mb[col_j-1]
         
          aug_adj_i[row_i, col_j] = adj[orig_pa, orig_ch]
      }
    }
    
    #set adjacencies of target node
    
    for(c_i in 1:(E_n-1)){
      row_i = node_n_i + c_i
      group_of_this_context = shape_to_part(shape)[c_i+1] #plus one as we skip first context as observational
      observational_group = shape_to_part(shape)[1]
      if(!(observational_group==group_of_this_context)) aug_adj_i[row_i, 1] = 1
    }
    node_adj[[paste0("y", node_i)]]  = aug_adj_i
    
    sys_adj_i = aug_adj_i[node_n_i,][node_n_i]
    
       
      
    if(write_data &
       length(mb)>0){ #needed if we run JCI afterwards
      write.csv(node_data_jci[[paste0("y", node_i)]],file=paste0("~/Vario/data_synthetic/", subdir_nm, seed, '-y', node_i, '-data.csv'),row.names=FALSE)
      write_metadata(Xy_n=node_n_i, E_n, C_n=E_n-1, D_n, fnm=paste0("~/Vario/data_synthetic/", subdir_nm, seed, '-y',  node_i), seed, node_i, shape)
      write.csv(aug_adj_i,file=paste0("~/Vario/out/graphs_true/", subdir_nm, E_n, "-", 
                                      seed, "-y", node_i, '-truegraph.csv' ),row.names=FALSE)
      # For easy comparison when analysing the results of relatedwork:
      write.csv(aug_adj_i,file=paste0("~/Vario/out_relatedwork/",  subdir_nm, 
                                      seed, "-y", node_i, '-truegraph.csv' ),row.names=FALSE) 
   
      }
  }
  
  
  return(list(node_data=node_data,node_data_jci=node_data_jci,
              node_MB=node_MB,
              node_adj=node_adj,
              adj=adj,
              empty_dag=FALSE))
}

# Invariant coefficients --------
#' Generate non-zero coefficients for \code{f:{X_1,...X_n} -> Y} uar from \code{[+-0.25,+-1]}
#' @param n num X
gen_DAG_weights = function(n){
  coefficients  = sapply(1:n,function(i) 
    return(c(sample_real(c(-1,-0.25),1), sample_real(c(0.25,1),1))[sample.int(2,1)]))
  return(coefficients)
}

#Changing coefficients -------
#' Given partition Pi, generate coefficients for \code{f^Pi: \{X_1,...X_n\}\} 
#' @details from \code{[+-0.5,+-1.5]}
#' over all contexts, e.g. for Xi and Pi=\{3,1}, 3 similar coefficients and one different (inter and intra group distances)
#' @param X_n num X 
#' @param shape pi, without gaps (linear coefs are totally ordered)
#' @param inter min distance between groups (to be rescaled wrt [-1.5,1.5]) 
gen_DAG_weights_Pi = function(X_n, shape, inter, noise, vb=FALSE, perfect_interventions=FALSE){  

   
  E_n = sum(shape)
assert(X_n>0)
cfs_C_X = matrix(0, sum(E_n), X_n)

for(X_i in 1:X_n){
# from [-1.5,-0.5] union [0.5,1.5]
first = sample_real(c(0.5,1.5),1)
if(sample.int(2,1)==2) first = -first

if(length(shape)==1) { #invariance
  cfs_C_X[,X_i] = rep(first, E_n)
  next
}
inc = (first<0)

rest= (abs(first)+1.5)
max_inter =  (rest/(length(shape)-1))-0.01 # this is how large intergroup distance may be, depending on how many groups fit into [-1,1]

inter[1] = inter[1]  *( max_inter/inter[2])
  inter[2] = max_inter
  
  if(inc) assert(first + (length(shape)-1)*max_inter <=1.5)
  else assert(first - (length(shape)-1)*max_inter >=-1.5)

  #the first group is observational and should be in [-1.5,-0.5] union [0.5,1.5]. other groups can be in [-1.5, 1.5], where inter ensures proper min. group distance.
#for(X_i in 1:X_n){
  cfs =c(first)
  for (grp in shape){
    if(grp>1){
      for (mem in 1:(grp-1))
       # if(inc) cfs = c(cfs, cfs[length(cfs)])
       # else 
        cfs = c(cfs, cfs[length(cfs)])
    }
    if(inc) cfs=c(cfs,cfs[length(cfs)]+sample_real(inter,1))
    else cfs=c(cfs,cfs[length(cfs)]-sample_real(inter,1))
  }
  for(i in 1:length(cfs)) if(sample.int(2,1)==1) cfs[i] = cfs[i] + sample_real(noise,1)
  else  cfs[i] = cfs[i] - sample_real(noise,1)
  cfs = cfs[-length(cfs)]  #at last grp, unnecessarily added member
  if(cfs[length(cfs)]==0) cfs = cfs+1
# no scaling needed when inter is rescaled  cfs =cfs * (2/cfs[length(cfs)]) -1  #rescale to [-1, 1]; can be zero for groups, = perf. intervention. 
  
  
  # exception: if there is only one group, steer away from zero (this is our observational group) 
  

    chosen = sample_real(c(0.25,1),1) 
  if(sample.int(2,1)==2) chosen = -chosen
    if(length(shape)==1) cfs = rep(chosen, length(cfs))
  
  cfs_C_X[,X_i] = cfs
}
  
  if(perfect_interventions){ #for simplicity, set the second group to zero and leave the rest as is - case e.g. shape = c(1,2,3,2) generates coefficients with usual spacing for the other groups.
    cfs_C_X = (gen_DAG_weights_interventions(cfs_C_X, X_n, shape, inter, vb))
    
  }  
  if(vb) apply(cfs_C_X,2, function(row) cat("\t", sapply(row, function(i) return(fmtd(i,2))),"\n"))

return(cfs_C_X)
}
# gen_DAG_weights_Pi_positive = function(X_n, shape, inter, intra, vb=FALSE, perfect_interventions=FALSE){
#   
#   if(perfect_interventions){ #for simplicity, set the second group to zero and leave the rest as is - case e.g. shape = c(1,2,3,2) generates coefficients with usual spacing for the other groups.
#     return(gen_DAG_weights_Pi(X_n, shape, inter, intra, vb))
#     
#   }
#   assert(X_n>0)
#   cfs_C_X = matrix(0, sum(shape), X_n)
#   for(X_i in 1:X_n){
#   cfs =c(0)
#   for (grp in shape){
#     if(grp>1){
#     for (mem in 1:(grp-1))
#       cfs = c(cfs, cfs[length(cfs)]+sample_real(intra,1))
#     }
#     cfs=c(cfs,cfs[length(cfs)]+sample_real(inter,1))
#   }
#    
#   cfs = cfs[-length(cfs)]  #at last grp, unnecessarily added member
#   if(cfs[length(cfs)]==0) cfs = cfs+1
#   cfs =cfs * (0.75/cfs[length(cfs)]) + 0.25  #rescale to [0.25,1]
#    
#   cfs_C_X[,X_i] = cfs
#   }
#    
#   if(vb) apply(cfs_C_X,2, function(row) cat("\t |", sapply(row, function(i) return(fmtd(i,2))),"\n"))
#   return(cfs_C_X)
# }

gen_DAG_weights_interventions = function(cfs_C_X, X_n, shape, inter,  vb=FALSE){
  intra=c(0,0) 
  
  for(X_i in 1:X_n){ 
    cfs= cfs_C_X[,X_i] 
    
    if(length(shape)>=2)  cfs[(shape[1]+1):(shape[1]+shape[2])] = 0
    else cfs = rep(0, sum(shape))

    cfs_C_X[,X_i] = cfs
  }
  return(cfs_C_X)
}



# quick test LM 
#x = data.frame(cbind(g$y6$D1$x1 ,cbind(g$y6$D1$x2 ,  g$y6$D1$x4  )))
#names(x) = c("x1", "x2", "x4")
#y= g$y6$D1$y
#data = data.frame(x,y)
#names(data) = c(names(x), "y")
#lm_mv(data, names(x), c(1,1,1))






get_MB = function(adj, y_i, vb=FALSE){
  if(vb) cat( "\n Y=", y_i,"\t") 
  
  #The original indices in Adj
  pa_i = which(adj[,y_i]==1)
  ch_i = which(adj[y_i,]==1)
  z_i = c() #parents of children
  for (child in ch_i) z_i = c(z_i, which(adj[,child]==1)) 
  z_i = z_i[-which(z_i==y_i)]
  z_i = unique(z_i) 
  
  MB_i = unique(c(c(pa_i, ch_i),z_i)) 
  
  

  pa_ind = which(MB_i %in% pa_i)
  ch_ind = which(MB_i %in% ch_i)
  z_ind = which(MB_i %in% z_i)
  non_ch_ind = unique(c(pa_ind, z_ind))
  
  
  overlap = which(non_ch_ind %in% ch_ind)
  if(length( overlap))
    non_ch_ind =non_ch_ind[-overlap]
   if (vb) cat("PA:",  pa_ind, 
      "\tCH:", ch_ind, "\tZ:",  z_ind, "NCH:", non_ch_ind, "\tNoisy: ")
  
  return(list(mb=MB_i, 
              pa_ind=pa_ind, ch_ind=ch_ind,  z_ind=z_ind, non_ch_ind=non_ch_ind,
              pa_n=length(pa_ind), ch_n=length(ch_ind), z_n=length(z_ind),non_ch_n=length(non_ch_ind)))
  #confusing:
  #return(list(pa_ind=pa_ind, #1:length
   #           ch_ind=ch_ind,
    #          z_ind=z_ind,
     #         non_ch_ind=non_ch_ind,
      #        mb=MB_i, # e.g. 2,4,5 if x2->Y and Y->4,5.
       #       mb_ind=1:length(MB_i)))
  # ind means re-ordered indices; mb has original indices; mb[pa_ind[1]] has for example orig. index of first parent; in adjacency matrix, put Y at position 1 and others +1 
  
}


# similar for Vario
dg2graphviz <- function(filename,p,
                        contextVars,
                        labels,dg) {
  sink(filename)
  cat('digraph G {\n');
  cat('subgraph cluster_system {\n');
  cat('label="system";\n')
  cat('style="invis";\n')
  if( p > 0 ) {
    for( i in 1:p )
      if( !(i %in% contextVars) )
        cat(i,'[label="',labels[i],'"];\n',sep='')
    for( i in 1:p )
      if( !(i %in% contextVars) ) {
        for( j in 1:p )
          if( !(j %in% contextVars) && i!=j )
            if( dg[i,j] ) {
              cat(i,'->',j,';\n',sep='')
            }
      }
  }
  cat('}\n')
  cat('subgraph context {\n');
  cat('label="context";\n')
  if( p > 0 ) {
    for( i in 1:p )
      if( i %in% contextVars )
        cat(i,'[label="',labels[i],'",shape=rectangle];\n',sep='')
    for( i in 1:p )
      if( i %in% contextVars ) {
        for( j in 1:p )
          if( !(j %in% contextVars) ) {
            if( dg[i,j] ) 
              cat(i,'->',j,';\n',sep='')
          }
      }
  }
  cat('}\n')
  cat('}\n')
  sink()
}


# Copyright (c) 2018-2020, Joris M. Mooij. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

dg2graphviz <- function(filename,p,contextVars,labels,dg) {
  sink(filename)
  cat('digraph G {\n');
  cat('subgraph cluster_system {\n');
  cat('label="system";\n')
  cat('style="invis";\n')
  if( p > 0 ) {
    for( i in 1:p )
      if( !(i %in% contextVars) )
        cat(i,'[label="',labels[i],'"];\n',sep='')
    for( i in 1:p )
      if( !(i %in% contextVars) ) {
        for( j in 1:p )
          if( !(j %in% contextVars) && i!=j )
            if( dg[i,j] ) {
              cat(i,'->',j,';\n',sep='')
            }
      }
  }
  cat('}\n')
  cat('subgraph context {\n');
  cat('label="context";\n')
  if( p > 0 ) {
    for( i in 1:p )
      if( i %in% contextVars )
        cat(i,'[label="',labels[i],'",shape=rectangle];\n',sep='')
    for( i in 1:p )
      if( i %in% contextVars ) {
        for( j in 1:p )
          if( !(j %in% contextVars) ) {
            if( dg[i,j] ) 
              cat(i,'->',j,';\n',sep='')
          }
      }
  }
  cat('}\n')
  cat('}\n')
  sink()
}


write_metadata = function(Xy_n, E_n, C_n, D_n, fnm, seed, node_i,shape){
  assert(C_n == E_n -1) #one observational "regime"
  #For jci, need metadata 
  metadata=list()
  metadata$SystemVars<- (1:Xy_n) # covariates and target y
  metadata$ContextVars<- ((Xy_n+1):(Xy_n+C_n))
  metadata$basefilename<-fnm
  metadata$pSysObs<-Xy_n
  metadata$pSysConf<-0 #no confounding
  metadata$pContext<-C_n
  metadata$eps<-0.5
  metadata$eta<-0.5
  metadata$N<-D_n * E_n #N*nRegimes
  metadata$acyclic<-1
  metadata$sufficient<-1
  metadata$surgical<-1
  metadata$seed<-seed
  metadata$Itargets<-list() # TODO Itargets
  metadata$targets<-list() # as.list(2^(targets-1))
  
  
  
  metadata$ytarget<-node_i # for file names 
  metadata$true_partition_group_sizes<-shape 
  metadata$true_observational_group_for_target<-shape[1]  #by convention, the first one and no other.
  cat(file=paste(fnm,'.json',sep=''),toJSON(metadata))
  
}



