#' VarioG for causal DAG discovery
#' Given data \code{\{X_e\}}, in each context, 
#' find causal parents and partition for each target Y, \code{Y = f_{PI}(pa(Y))+N}  
#' then find a DAG, removing ancestral paths
#' @param D list: \code{C -> \{X_e\}} for each context C
#' @param X_max_n max. length of subsets of X to consider. default is 4.
#' @param it_n alt. to max. number of subsets, restrict iterations
#' @param it_pre rounds of precomputing to find confidence intervals 
#' @param confints (opt.) if precomputed already
#' @param X_names (for display) covariates X names
#' @param target_nm (for display) target Y name  
#' @param vb verbose
#' @param greedy_k greedy pi search
#' @return \code{resultVario} class summarizing results
#' @export
Vario_G_search = function(D,
                          X_names,
                          X_max_n=4,
                          it_n=500,
                          it_pre=50, 
                          greedy_k=FALSE, vb=TRUE,
                          target_nm="y",
                          confints = NULL,
                          version_mdl = TRUE,
                          version_simp = TRUE
                               ){ 
  #Phase 1: Vario for each target Y --------
  
  #D= load_Sachs_data()
  #names(D$E8)[4]="PIP2"
  #names(D$E8)[5]="PIP3"
  names(D)= sapply(1:length(D), function(i) 
    return(paste0("E", i)))  
  X_names = names(D[[1]])
  E_n = length(D)
  
 
  if(vb) cat("\n(varioG) init ... ")
  if(vb) tic(paste0("\n(varioG) ", it_pre, " calls", sep=""))
  confints  = list(conf_pione=sample_conf_pi_one(E_n, it_pre, FALSE,version_mdl = version_mdl),  
                  conf_nogroups= sample_conf_nogroups(E_n, it_pre, FALSE,version_mdl = version_mdl),
                  conf_best=sample_conf_pi_best(E_n, it_pre, FALSE,version_mdl = version_mdl)) 
  if(vb) toc()  
  
  edges_per_Y = list()
  results_per_Y = list()
  
  if(vb) cat("\n(varioG) Edge search for each target Y ")
  for (y_i in 1:length(X_names)){
    y_nm = X_names[y_i]
    if(vb) cat("\n(varioG) Y =", y_i, ":" , y_nm, "\n")
    xs = (1:length(D[[1]]))[-y_i]
    Dy = transform_xls_data(D, xs, c(y_i))
    D1_Dn = Dy$D_data
    x_nms = Dy$X_names[which(Dy$X_names != "y")]  

    
    RESULT = Vario_subset_search(D1_Dn, x_nms,  
                                 X_max_n=X_max_n,
                                 confints = confints,
                                 vb=vb, version_mdl=version_mdl,
                                 version_simp=version_simp,
                                 greedy_k=greedy_k)
    results_per_Y[[y_i]] = RESULT
    edges_per_Y[[y_i]] = RESULT$X_Pi_ranking # each triple (X,Y, Pi) found by Vario
  }
  
   
  #Phase 2: Pruning edges --------
  
  edges_ordered = get_edge_list(edges_per_Y,results_per_Y, X_names, confints, consider_conf = FALSE, vb = vb)
  edges_admissible_confident = get_edge_list(edges_per_Y,results_per_Y, X_names, confints, consider_conf =TRUE, vb=vb)
  
 
  result_pruned =   get_edges_nonancestral (edges_ordered, consider_3anc = TRUE, X_names, consider_conf_afterwards = FALSE)
  edges_pruned = result_pruned[[1]]
  
  return(list(edges_pruned = edges_pruned, edges_admissible = edges_ordered, edges_admissible_confident = edges_admissible_confident))
}

get_edge_list = function(results, 
                         result_info,
                         xi_names,  
                         confints_8, 
                         consider_conf=FALSE, 
                         consider_nogroup=TRUE, #todo should always be true
                         vb = TRUE){
  node_to_ingoing = list()
  edges_ordered = list()
  node_n  = length(results)
  for (node_i in 1:node_n){
    nm_i = xi_names[node_i]
    node_to_ingoing[[nm_i]] = list() #points to ingoing edges
    
    if(vb) cat("\n---------------------\nNode:", xi_names[node_i],  " ", node_i)
    result_i = as.list(results[[node_i]])
    for (edge_i in 1:length(result_i)){
      edge  = result_i[[edge_i]]
      
      pi = edge$E_part
      score=  edge$score_norm 
      
      k = length(unique(pi))
      # conf = mean(confints_8$conf_pione[,(k-1)])
      #conf = mean(confints_8$conf_nogroups[,(k-1)])
      conf = mean(confints_8$conf_pione[,(k-1)])
      if(consider_nogroup){
        if(length(unique(pi))==1) next #break 
        if(length(unique(pi))==length(pi)) next #break
      }
      else {
        if(length(unique(pi))==1)  break 
        if(length(unique(pi))==length(pi))  break
        if(length(unique(pi))>2) next
      }
      if(consider_conf & (score>conf)) next
      
      #if(match_partitions & !(length(unique(pi))==2)) next
      nm_j = edge$xnm
      node_j = which(xi_names==nm_j)
      node_to_ingoing[[nm_i]][[nm_j]] =  edge
      
      edge$nm_i = nm_i
      edge$nm_j = nm_j
      edge$node_i = node_i
      edge$node_j = node_j
      edge$pruned = FALSE
      nm = paste0(names(result_i)[edge_i], "_", nm_i)
      
      edges_ordered[[nm]] = edge
      
      score = edge$score_norm 
      
      xi_nm = edge$xnm
      xi = which(xi_names==xi_nm)  
      if(vb)  cat("\n\tEdge: ", xi_names[xi]," ", xi, paste0(pi, collapse=""))
      
      #if(adj[xi, node_i]==0) #do not overwrite edges with lower scores 
      #  adj[xi, node_i]= score
      #adj_bin[xi, node_i] = 1 
    }
  }
  edges_ordered = list.sort(edges_ordered, score_norm)
 
  for(i in 1:length(edges_ordered)){
      gain = edgeanalysis_MDL_gain(xi_names, edges_ordered, result_info, 
                                   edges_ordered[[i]], edges_ordered[[i]]$node_j, edges_ordered[[i]]$node_i)$gain
      edges_ordered[[i]]$gain = gain
  }
  
  return(edges_ordered)
}




get_edges_nonancestral =  function(edges_ordered,  
                           xi_names,  
                           match_partitions = TRUE,  
                           consider_3anc = TRUE,
                           consider_2anc = TRUE, 
                           consider_conf_afterwards= FALSE,
                           vb =FALSE){
  
   
  keepers = list()
  discard = list()
  
  adj_pruned_values = matrix(0, 11, 11)
  #anc_adj = matrix(0, 11, 11)
  
  if(vb)  cat("\nAncestral edges\n---------------------")
  FP_but_ANC = 0
  for (e in 1:length(edges_ordered)){
    edge_ji = edges_ordered[[e]]
    node_i = edge_ji$node_i
    node_j = edge_ji$node_j 
    
    if(vb)  cat("\nChecking ", xi_names[node_j], "->", xi_names[node_i])

    
    #for (nk in 1:11){
    #  if(nk==node_j | nk==node_i) next
    #  if( (adj_small[node_j,nk]==1 & adj_small[nk,node_i]==1)) {
    #    if(adj_small[node_j,node_i]==0 & adj_small[node_i,node_j]==0 ){
    #      edge_ji$exists_only_ancestral_small =TRUE
    #      anc_adj[node_i,node_j]= 1
    ##    }
   #     true_ancestors = paste0(true_ancestors, xi_names[nk], ",")
    #    edge_ji$exists_ancestral_small= TRUE
        
        
     # }
    #  if( (adj_aug[node_j,nk]==1 & adj_aug[nk,node_i]==1)) {
    #    if(adj_aug[node_j,node_i]==0 & adj_aug[node_i,node_j]==0){
    #      edge_ji$exists_only_ancestral_aug =TRUE
    #      anc_adj[node_i,node_j]= 1
    #    } 
    #    edge_ji$exists_ancestral_aug= TRUE
        
    #  }
     # for (nr in 1:11){
    #    if(nr==node_j | nr==nk | nr==node_i) next
     #   if( (adj_big[node_j,nr]==1 &adj_big[nr,nk]==1 & adj_big[nk,node_i]==1) &
    #        adj_big[node_j,node_i]==0 & adj_big[node_i,node_j]==0 ) 
    #      edge_ji$exists_only_3ancestral_aug= TRUE
    #    if( (adj_small[node_j,nr]==1 &adj_small[nr,nk]==1 & adj_small[nk,node_i]==1) & 
    #        adj_small[node_j,node_i]==0 & adj_small[node_i,node_j]==0 ) 
    #      edge_ji$exists_only_3ancestral_small= TRUE
     # }
      
      
   # }
 #   edge_ji$causal_small = (adj_small[node_j,node_i]==1)
 #   edge_ji$causal_aug = (adj_big[node_j,node_i]==1)
    
    #exists_ancestral = exists_ancestral | (adj_true[node_j,node_k]==1 & adj_true[node_k,node_i]==1)
  #  edge_ji$noncausal_small = (adj_small[node_j,node_i]==0)
  #  edge_ji$exists_anti_small = (adj_small[node_i,node_j]==1)
  #  edge_ji$exists_anti_big = (adj_big[node_i,node_j]==1)
    
    for (e2 in 1:length(edges_ordered)){ 
      edge_jk = edges_ordered[[e2]]
      node_jj = edge_jk$node_j
      node_k = edge_jk$node_i
      
      if(node_jj != node_j | node_k == node_i  | node_k == node_j ) next
      
      if(edge_jk$pruned | edge_ji$pruned) next
      
      
      for (e3 in 1:length(edges_ordered)){ 
        edge_ki = edges_ordered[[e3]]
        node_kk = edge_ki$node_j
        node_ii = edge_ki$node_i
        
        if(!(node_kk == node_k & node_ii == node_i)) next
        
        if(edge_jk$pruned | edge_ji$pruned | edge_ki$pruned) next
        
        
        if(vb)  cat("\n\tPath ", xi_names[node_j], "->", xi_names[node_k], "->", xi_names[node_i], 
                    ", Pis", paste0(edge_jk$E_part, collapse=""),"->" ,paste0(edge_ki$E_part, collapse=""),"->" , paste0(edge_ji$E_part, collapse=""))
        
        
        
        if(consider_2anc){
          # Look for n_j -> n_k -> n_i where Vario found: jk, ki, and ji.
          # Discard the edge nj, ni if there is at least one such path with better score (suggesting that ki is the true causal edge)
          # Discarding only works if the partitions at edge ki and ji agree (up to more-finegrained, i.e. are consistent) that is, they find the same "interventions" for node ni.
          ###   for (e_k in edges_k){    
          
          # n_j to n_k
          #may be multiple edges btw the same node pair!!!  looking for a smaller ancestral scre to reject-> can use the first entry
          
          
          #rji = MDL_cost_edge(xi_names, edges_ordered, results, 
          #                   edge_ji, node_j,  node_i)
          #rjk = MDL_cost_edge(xi_names, edges_ordered, results, 
          #                   edge_jk, node_j,  node_k)
          # rki = MDL_cost_edge(xi_names, edges_ordered, results, 
          #                    edge_ki, node_k,  node_i)
          #anc = rjk$diff + rki$diff
          anc = (edge_jk$score_norm + edge_ki$score_norm)
          anc = anc/2
          dir = edge_ji$score_norm
          
          pi1 = edge_jk$E_part
          pi2 = edge_ki$E_part
          pi3 = edge_ji$E_part 
          
          consistent  = (check_finegrained(pi2,pi3)$finegrained | check_finegrained(pi3,pi2)$finegrained) 
         
          if( (anc < dir)   #At least one ancestral path (via z) is found that is better (cheaper by score) than the direct one
              & (consistent | (!match_partitions)))  { 
            
            edge_ji$pruned = TRUE
            if(vb)  cat("\n\t Removed:",    xi_names[node_j],  "->",xi_names[node_k] , "->", xi_names[node_i], "\t\t", anc,"<", dir, "\t CONS:", consistent)
            
            adj_pruned_values[node_j,node_i] = anc-dir
          }
          #TODO
          # #jrki 
          if(consider_3anc)  
            for (e4 in 1:length(edges_ordered))
              for (e5 in 1:length(edges_ordered)){
                edge_jr = edges_ordered[[e4]]
                edge_rk = edges_ordered[[e5]]
                
                node_jj = edge_jr$node_j
                node_r= edge_jr$node_i
                node_rr= edge_rk$node_j
                node_kk= edge_rk$node_i
                if(node_r == node_k | node_r == node_j | node_r==node_i | 
                   node_r != node_rr | node_k != node_kk | node_j != node_jj) next
                
                
                if(vb) cat(" \n\t | Path: ", xi_names[node_j], "->",  xi_names[node_r], "->", xi_names[node_k], "->", xi_names[node_i])
                
                
                
                anc = (edge_jr$score_norm + edge_rk$score_norm + edge_ki$score_norm)/3
                dir = edge_ji$score_norm
                
                if(anc < dir)  {  
                  consistent = "TODO"
                  edge_ji$pruned = TRUE
                  if(vb)  cat("\n\t Removed:",    xi_names[node_j],  "->",xi_names[node_k] , "->", xi_names[node_i], "\t\t", anc,"<", dir, "\t CONS:", consistent)
                  
                  
                  adj_pruned_values[node_j,node_i] = anc-dir
                }
              }
          
          
        } 
      }
    } 
    
    #All edges k processed at this point
    edge_ji$node_pa = node_j
    edge_ji$node_child = node_i
    #edge_ji$ancestral= exists_ancestral
    #edge_ji$true_ancestors= true_ancestors
    #edge_ji$causal= causal 
    
    #if(exists_only_ancestral) anc_adj[node_j,node_i]= 1
    if(!edge_ji$pruned){ 
      if(vb) cat("\n\t Keeping:",xi_names[node_j], "-> {",true_ancestors,
                 " } ->", xi_names[node_i])#, "\t\t", anc, "<",adj_to_i[x]) 
      
    }
    
    if(vb) cat("\n---------------------\n")
    
    if(consider_conf_afterwards) {
      pi = edge_ji$E_part
      k = length(unique(pi))
      conf =  42.85396#mean(confints_8$conf_pione[,(k-1)])
      score= edge_ji$score_norm
      if(score>conf) edge_ji$pruned=  TRUE
      #     if(noncausal) fp = "++ noncausal\t"
      #     if(causal) fp = "-- causal\t" 
      #     if(vb)  cat("\n\t Removed:", fp, xi_names[node_j],  "->",xi_names[node_i], "by conf")
      #     pruned =TRUE
      #   }
      # }
    } 
    
    edges_ordered[[e]] = edge_ji
 
  }
  for(e in edges_ordered){
    nm = paste0(xi_names[node_j], "->",xi_names[node_i])
    if(!e$pruned)  keepers[[nm]] = e
    else discard[[nm]] = e
  }
  
  
   
  found = matrix(0,11,11)
  #remove dups
  edges_pruned = list()
  for(e in edges_ordered){
    if(!e$pruned & found[e$node_j, e$node_i] == 0){
      found[e$node_j, e$node_i] = 1
      
      nm = paste0(xi_names[e$node_j], "_" , paste0(e$E_part,collapse = ""), "_",
                  xi_names[e$node_i] )
      edges_pruned[[nm]] = e
    }
  }
  
  adj_bin_pruned = matrix(0,11,11)
  for(e in edges_pruned) if(!e$pruned) adj_bin_pruned[e$node_j, e$node_i]=1
  
  return(list(edges_pruned, adj_bin_pruned, adj_pruned_values))
  
}


get_counts_ground_truth = function(edges_pruned, adj_anc, adj_anc3, adj_truth, X_names) {# = consensus_augmented_graph_Sachs(X_names)
  E_n = length(X_names)
  edges_pruned = annotate_edges(edges_pruned)
  adj_bin_pruned = matrix(0,E_n,E_n)
    
  
  #True positives
  for(e in edges_pruned) if(!e$pruned) 
    adj_bin_pruned[e$node_j, e$node_i]=1 
  
  # extracting ancestral edges for Sachs
  #adjs = consensus_ancestral_graph(adj_truth)
  #adj_anc = adjs[[1]]
  #adj_anc3 = adjs[[2]]
  

  
  adj_guess = adj_bin_pruned
  
  tp = length(which(adj_guess + adj_truth == 2))
  fp = length(which(adj_truth - adj_guess == -1))
  fn = length(which(adj_guess - adj_truth == -1)) 
  
  
  bi =  length(which(t(adj_truth) + adj_guess + adj_truth == 3)) 
  ac = length(which(t(adj_truth) + adj_guess == 2)) 
  ac = ac - bi
  
  
  adjs = consensus_ancestral_graph(adj_truth)
  adj_anc = adjs[[1]]
  adj_anc3 = adjs[[2]]
  tp_anc = length(which(adj_guess + adj_anc == 2)) 
  tp_anc3 = length(which(adj_guess + adj_anc3 == 2))
  
  
  
  assert(length(which(adj_anc3 + adj_anc==2)) ==0) #dont count edges that are 2anc as 3anc
  
  assert(length(which(adj_guess + adj_truth + adj_anc==3)) ==0) #no causal edges are in anc
  assert(length(which(adj_guess + adj_truth + adj_anc3==3)) ==0) 
  assert(length(which(t(adj_guess) + adj_truth + adj_anc ==3)) ==0)
  assert(length(which(t(adj_guess) + adj_truth + adj_anc3 ==3)) ==0) # no noncausal edges are in anc
  
  
  
  #Printing ---------------
  for ( i in 1:E_n) 
     for (j in 1:E_n){
       if(adj_guess[i,j]==1){
         if(adj_anc[i,j]==1) ok = "ancestral \t (a)"
         else{
           if(adj_anc3[i,j]==1) ok = "3-ancestral \t (a)"
          else{
            if(adj_truth[i,j]==1) ok = "causal \t (+)"
            else if(adj_truth[j,i]==1) ok = "anticausal \t (-)" 
            else ok = "no edge \t ()"
          }
         }
         cat("\n\t", X_names[i], "\t->\t", X_names[j], "\t"  ,ok) 
       }
         
} 
  
  f1 = tp/(tp + (1/2)* (fp + fn))
  f1good = tp/(tp + (1/2)* ((fp-(tp_anc+tp_anc3)) + fn))
  cat("\nResults: \n\tTP   / FP (thereof: A / A3 / R )", " / missed = f1 \n")
  cat("\t ",tp,   "/ ",    fp, "(", tp_anc, tp_anc3, ac,    ")", "/", fn, "=",f1, "\n")
  cat("\nMeaning: A = ancestral but not directly causal or anticausal \n\t A3 = ancestral over 2 nodes but not directly ancestral,causal,anticausal \n\t R = anticausal (and not bidirectional) ")
  
  
  return(list(counts = c(tp,fp,f1), adj = adj_bin_pruned))
}


annotate_edges = function(edges , vb = FALSE){
  
  D= load_Sachs_data()
  names(D$E8)[4]="PIP2"
  names(D$E8)[5]="PIP3"
  names(D)= sapply(1:length(D), function(i) 
    return(paste0("E", i)))  
  xi_names = names(D[[1]])
  
  
  adj_big = consensus_augmented_graph_Sachs(xi_names)
  adj_aug = adj_big
  adj_small = consensus_small_graph_Sachs(xi_names)
  for (e in 1:length(edges)){
    edge_ji = edges[[e]]
    node_i = edge_ji$node_i
    node_j = edge_ji$node_j
    #exists_ancestral = FALSE 
    #exists_only_ancestral =FALSE
    true_ancestors = ""
    if(vb)  cat("\nChecking ", xi_names[node_j], "->", xi_names[node_i])
    edge_ji$exists_ancestral_small= FALSE 
    edge_ji$exists_ancestral_aug= FALSE
    
    edge_ji$exists_only_ancestral_small= FALSE 
    edge_ji$exists_only_ancestral_aug= FALSE
    edge_ji$exists_only_3ancestral_small= FALSE 
    edge_ji$exists_only_3ancestral_aug= FALSE
    
    for (nk in 1:11){
      if(nk==node_j | nk==node_i) next
      if( (adj_small[node_j,nk]==1 & adj_small[nk,node_i]==1)) {
        if(adj_small[node_j,node_i]==0 & adj_small[node_i,node_j]==0 ){
          edge_ji$exists_only_ancestral_small =TRUE
         # anc_adj[node_i,node_j]= 1
        }
        true_ancestors = paste0(true_ancestors, xi_names[nk], ",")
        edge_ji$exists_ancestral_small= TRUE
        
        
      }
      if( (adj_aug[node_j,nk]==1 & adj_aug[nk,node_i]==1)) {
        if(adj_aug[node_j,node_i]==0 & adj_aug[node_i,node_j]==0){
          edge_ji$exists_only_ancestral_aug =TRUE
       #   anc_adj[node_i,node_j]= 1
        } 
        edge_ji$exists_ancestral_aug= TRUE
        
      }
      for (nr in 1:11){
        if(nr==node_j | nr==nk | nr==node_i) next
        if( (adj_big[node_j,nr]==1 &adj_big[nr,nk]==1 & adj_big[nk,node_i]==1) &
            adj_big[node_j,node_i]==0 & adj_big[node_i,node_j]==0 ) 
          edge_ji$exists_only_3ancestral_aug= TRUE
        if( (adj_small[node_j,nr]==1 &adj_small[nr,nk]==1 & adj_small[nk,node_i]==1) & 
            adj_small[node_j,node_i]==0 & adj_small[node_i,node_j]==0 ) 
          edge_ji$exists_only_3ancestral_small= TRUE
      }
      
      
    }
    edge_ji$causal_small = (adj_small[node_j,node_i]==1)
    edge_ji$causal_aug = (adj_big[node_j,node_i]==1)
    
    #exists_ancestral = exists_ancestral | (adj_true[node_j,node_k]==1 & adj_true[node_k,node_i]==1)
    edge_ji$noncausal_small = (adj_small[node_j,node_i]==0)
    edge_ji$exists_anti_small = (adj_small[node_i,node_j]==1)
    edge_ji$exists_anti_big = (adj_big[node_i,node_j]==1)
    
     
     edges[[e]] = edge_ji
  }
  
  return(edges)
}